@@ -27,6 +27,12 @@
 /* for sig_atomic_t */
 #include <signal.h>
 
+#ifdef HAVE_FUNC_ATTRIBUTE_FALLTHROUGH
+#define MC_FALLTHROUGH __attribute__((fallthrough))
+#else
+#define MC_FALLTHROUGH
+#endif
+
 /*** typedefs(not structures) and defined constants **********************************************/
 
 /* The O_BINARY definition was taken from gettext */
@@ -128,27 +134,13 @@
 #define TMPDIR_DEFAULT "/tmp"
 #define SCRIPT_SUFFIX ""
 #define get_default_editor() "vi"
-#define OS_SORT_CASE_SENSITIVE_DEFAULT 1
+#define OS_SORT_CASE_SENSITIVE_DEFAULT TRUE
 #define UTF8_CHAR_LEN 6
 
 /* Used to distinguish between a normal MC termination and */
 /* one caused by typing 'exit' or 'logout' in the subshell */
 #define SUBSHELL_EXIT 128
 
-/* C++ style type casts */
-#define const_cast(m_type, m_expr) ((m_type) (m_expr))
-
-#if 0
-#ifdef MC_ENABLE_DEBUGGING_CODE
-#undef NDEBUG
-#else
-#define NDEBUG
-#endif
-#ifdef HAVE_ASSERT_H
-#include <assert.h>
-#endif
-#endif
-
 #define MC_ERROR g_quark_from_static_string (PACKAGE)
 
 #define DEFAULT_CHARSET "ASCII"
@@ -191,12 +183,12 @@ typedef struct
     int display_codepage;
 #else
     /* If true, allow characters in the range 160-255 */
-    int eight_bit_clean;
+    gboolean eight_bit_clean;
     /*
      * If true, also allow characters in the range 128-159.
      * This is reported to break on many terminals (xterm, qansi-m).
      */
-    int full_eight_bits;
+    gboolean full_eight_bits;
 #endif                          /* !HAVE_CHARSET */
     /*
      * If utf-8 terminal utf8_display = TRUE
