@@ -2,7 +2,7 @@
    Interface to the terminal controlling library.
    Ncurses wrapper.
 
-   Copyright (C) 2005-2016
+   Copyright (C) 2005-2018
    Free Software Foundation, Inc.
 
    Written by:
@@ -80,7 +80,7 @@
 /*** file scope variables ************************************************************************/
 
 /* ncurses supports cursor positions only within window */
-/* We use our own cursor coordibates to support partially visible widgets */
+/* We use our own cursor coordinates to support partially visible widgets */
 static int mc_curs_row, mc_curs_col;
 
 /*** file scope functions ************************************************************************/
@@ -179,6 +179,8 @@ mc_tty_normalize_lines_char (const char *ch)
 void
 tty_init (gboolean mouse_enable, gboolean is_xterm)
 {
+    struct termios mode;
+
     initscr ();
 
 #ifdef HAVE_ESCDELAY
@@ -194,11 +196,15 @@ tty_init (gboolean mouse_enable, gboolean is_xterm)
     ESCDELAY = 200;
 #endif /* HAVE_ESCDELAY */
 
+    tcgetattr (STDIN_FILENO, &mode);
     /* use Ctrl-g to generate SIGINT */
-    cur_term->Nttyb.c_cc[VINTR] = CTRL ('g');   /* ^g */
+    mode.c_cc[VINTR] = CTRL ('g');      /* ^g */
     /* disable SIGQUIT to allow use Ctrl-\ key */
-    cur_term->Nttyb.c_cc[VQUIT] = NULL_VALUE;
-    tcsetattr (cur_term->Filedes, TCSANOW, &cur_term->Nttyb);
+    mode.c_cc[VQUIT] = NULL_VALUE;
+    tcsetattr (STDIN_FILENO, TCSANOW, &mode);
+
+    /* curses remembers the "in-program" modes after this call */
+    def_prog_mode ();
 
     tty_start_interrupt_key ();
 
@@ -219,8 +225,6 @@ tty_init (gboolean mouse_enable, gboolean is_xterm)
 void
 tty_shutdown (void)
 {
-    disable_mouse ();
-    disable_bracketed_paste ();
     tty_reset_shell_mode ();
     tty_noraw_mode ();
     tty_keypad (FALSE);
