@@ -1,7 +1,7 @@
 /*
    Virtual File System switch code
 
-   Copyright (C) 1995-2016
+   Copyright (C) 1995-2018
    Free Software Foundation, Inc.
 
    Written by: 1995 Miguel de Icaza
@@ -79,8 +79,6 @@ vfs_class *current_vfs = NULL;
 
 #define VFS_FIRST_HANDLE 100
 
-#define ISSLASH(a) (a == '\0' || IS_PATH_SEP (a))
-
 /*** file scope type declarations ****************************************************************/
 
 struct vfs_openfile
@@ -95,7 +93,7 @@ struct vfs_openfile
 /** They keep track of the current directory */
 static vfs_path_t *current_path = NULL;
 
-static GPtrArray *vfs_openfiles;
+static GPtrArray *vfs_openfiles = NULL;
 static long vfs_free_handle_list = -1;
 
 /* --------------------------------------------------------------------------------------------- */
@@ -202,6 +200,19 @@ vfs_get_openfile (int handle)
     return h;
 }
 
+/* --------------------------------------------------------------------------------------------- */
+
+static gboolean
+vfs_test_current_dir (const vfs_path_t * vpath)
+{
+    struct stat my_stat, my_stat2;
+
+    return (mc_global.vfs.cd_symlinks && mc_stat (vpath, &my_stat) == 0
+            && mc_stat (vfs_get_raw_current_dir (), &my_stat2) == 0
+            && my_stat.st_ino == my_stat2.st_ino && my_stat.st_dev == my_stat2.st_dev);
+}
+
+
 /* --------------------------------------------------------------------------------------------- */
 /*** public functions ****************************************************************************/
 /* --------------------------------------------------------------------------------------------- */
@@ -308,24 +319,24 @@ vfs_register_class (struct vfs_class * vfs)
 char *
 vfs_strip_suffix_from_filename (const char *filename)
 {
-    char *semi, *p, *vfs_prefix;
+    char *semi, *p;
 
     if (filename == NULL)
         vfs_die ("vfs_strip_suffix_from_path got NULL: impossible");
 
     p = g_strdup (filename);
     semi = g_strrstr (p, VFS_PATH_URL_DELIMITER);
-    if (semi == NULL)
-        return p;
-
-    *semi = '\0';
-    vfs_prefix = strrchr (p, PATH_SEP);
-    if (vfs_prefix == NULL)
+    if (semi != NULL)
     {
-        *semi = *VFS_PATH_URL_DELIMITER;
-        return p;
+        char *vfs_prefix;
+
+        *semi = '\0';
+        vfs_prefix = strrchr (p, PATH_SEP);
+        if (vfs_prefix == NULL)
+            *semi = *VFS_PATH_URL_DELIMITER;
+        else
+            *vfs_prefix = '\0';
     }
-    *vfs_prefix = '\0';
 
     return p;
 }
@@ -482,10 +493,16 @@ vfs_shut (void)
             vfs->done (vfs);
     }
 
+    /* NULL-ize pointers to make unit tests happy */
     g_ptr_array_free (vfs_openfiles, TRUE);
+    vfs_openfiles = NULL;
     g_ptr_array_free (vfs__classes_list, TRUE);
+    vfs__classes_list = NULL;
     g_string_free (vfs_str_buffer, TRUE);
-    g_free (mc_readdir_result);
+    vfs_str_buffer = NULL;
+    current_vfs = NULL;
+    vfs_free_handle_list = -1;
+    MC_PTR_FREE (mc_readdir_result);
 }
 
 /* --------------------------------------------------------------------------------------------- */
@@ -554,12 +571,7 @@ vfs_setup_cwd (void)
 
         if (tmp_vpath != NULL)
         {
-            struct stat my_stat, my_stat2;
-
-            if (mc_global.vfs.cd_symlinks
-                && mc_stat (tmp_vpath, &my_stat) == 0
-                && mc_stat (vfs_get_raw_current_dir (), &my_stat2) == 0
-                && my_stat.st_ino == my_stat2.st_ino && my_stat.st_dev == my_stat2.st_dev)
+            if (vfs_test_current_dir (tmp_vpath))
                 vfs_set_raw_current_dir (tmp_vpath);
             else
                 vfs_path_free (tmp_vpath);
@@ -575,14 +587,11 @@ vfs_setup_cwd (void)
         g_free (current_dir);
 
         if (tmp_vpath != NULL)
-        {                       /* One of the directories in the path is not readable */
-            struct stat my_stat, my_stat2;
+        {
+            /* One of directories in the path is not readable */
 
             /* Check if it is O.K. to use the current_dir */
-            if (!(mc_global.vfs.cd_symlinks
-                  && mc_stat (tmp_vpath, &my_stat) == 0
-                  && mc_stat (vfs_get_raw_current_dir (), &my_stat2) == 0
-                  && my_stat.st_ino == my_stat2.st_ino && my_stat.st_dev == my_stat2.st_dev))
+            if (!vfs_test_current_dir (tmp_vpath))
                 vfs_set_raw_current_dir (tmp_vpath);
             else
                 vfs_path_free (tmp_vpath);
