@@ -2,7 +2,7 @@
 
 /* Background support.
 
-   Copyright (C) 1996-2015
+   Copyright (C) 1996-2018
    Free Software Foundation, Inc.
 
    Written by:
@@ -42,7 +42,6 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/wait.h>           /* waitpid() */
-#include <fcntl.h>
 
 #include "lib/global.h"
 
@@ -77,7 +76,7 @@ static int parent_fd;
 /* File descriptor for messages from our parent */
 static int from_parent_fd;
 
-struct TaskList *task_list = NULL;
+TaskList *task_list = NULL;
 
 static int background_attention (int fd, void *closure);
 
@@ -200,7 +199,7 @@ background_attention (int fd, void *closure)
         int (*non_have_ctx3) (file_op_context_t *, int, char *, char *, char *);
         int (*non_have_ctx4) (file_op_context_t *, int, char *, char *, char *, char *);
 
-        char *(*ret_str0) ();
+        char *(*ret_str0) (void);
         char *(*ret_str1) (char *);
         char *(*ret_str2) (char *, char *);
         char *(*ret_str3) (char *, char *, char *);
@@ -212,23 +211,22 @@ background_attention (int fd, void *closure)
     int argc, i, status;
     char *data[MAXCALLARGS];
     ssize_t bytes, ret;
-    struct TaskList *p;
+    TaskList *p;
     int to_child_fd = -1;
     enum ReturnType type;
+    const char *background_process_error = _("Background process error");
 
     ctx = closure;
 
     bytes = read (fd, &routine.pointer, sizeof (routine));
     if (bytes == -1 || (size_t) bytes < (sizeof (routine)))
     {
-        const char *background_process_error = _("Background process error");
-
         unregister_task_running (ctx->pid, fd);
         if (!waitpid (ctx->pid, &status, WNOHANG))
         {
             /* the process is still running, but it misbehaves - kill it */
             kill (ctx->pid, SIGTERM);
-            message (D_ERROR, background_process_error, _("Unknown error in child"));
+            message (D_ERROR, background_process_error, "%s", _("Unknown error in child"));
             return 0;
         }
 
@@ -236,7 +234,7 @@ background_attention (int fd, void *closure)
         if (WIFEXITED (status) && (WEXITSTATUS (status) == 0))
             return 0;
 
-        message (D_ERROR, background_process_error, _("Child died unexpectedly"));
+        message (D_ERROR, background_process_error, "%s", _("Child died unexpectedly"));
 
         return 0;
     }
@@ -250,14 +248,14 @@ background_attention (int fd, void *closure)
 
     if (argc > MAXCALLARGS)
     {
-        message (D_ERROR, _("Background protocol error"),
+        message (D_ERROR, _("Background protocol error"), "%s",
                  _("Background process sent us a request for more arguments\n"
                    "than we can handle."));
     }
 
     if (have_ctx)
     {
-        if (read (fd, ctx, sizeof (file_op_context_t)) != sizeof (file_op_context_t))
+        if (read (fd, ctx, sizeof (*ctx)) != sizeof (*ctx))
         {
             return reading_failed (-1, data);
         }
@@ -291,7 +289,7 @@ background_attention (int fd, void *closure)
         to_child_fd = p->to_child_fd;
 
     if (to_child_fd == -1)
-        message (D_ERROR, _("Background process error"), _("Unknown error in child"));
+        message (D_ERROR, background_process_error, "%s", _("Unknown error in child"));
 
     /* Handle the call */
     if (type == Return_Integer)
@@ -316,6 +314,8 @@ background_attention (int fd, void *closure)
             case 4:
                 result = routine.have_ctx4 (Background, data[0], data[1], data[2], data[3]);
                 break;
+            default:
+                break;
             }
         else
             switch (argc)
@@ -336,12 +336,14 @@ background_attention (int fd, void *closure)
                 result =
                     routine.non_have_ctx4 (ctx, Background, data[0], data[1], data[2], data[3]);
                 break;
+            default:
+                break;
             }
 
         /* Send the result code and the value for shared variables */
-        ret = write (to_child_fd, &result, sizeof (int));
+        ret = write (to_child_fd, &result, sizeof (result));
         if (have_ctx && to_child_fd != -1)
-            ret = write (to_child_fd, ctx, sizeof (file_op_context_t));
+            ret = write (to_child_fd, ctx, sizeof (*ctx));
     }
     else if (type == Return_String)
     {
@@ -413,12 +415,12 @@ parent_call_header (void *routine, int argc, enum ReturnType type, file_op_conte
     have_ctx = (ctx != NULL);
 
     ret = write (parent_fd, &routine, sizeof (routine));
-    ret = write (parent_fd, &argc, sizeof (int));
+    ret = write (parent_fd, &argc, sizeof (argc));
     ret = write (parent_fd, &type, sizeof (type));
     ret = write (parent_fd, &have_ctx, sizeof (have_ctx));
 
     if (have_ctx)
-        ret = write (parent_fd, ctx, sizeof (file_op_context_t));
+        ret = write (parent_fd, ctx, sizeof (*ctx));
     (void) ret;
 }
 
@@ -439,13 +441,13 @@ parent_va_call (void *routine, gpointer data, int argc, va_list ap)
 
         len = va_arg (ap, int);
         value = va_arg (ap, void *);
-        ret = write (parent_fd, &len, sizeof (int));
+        ret = write (parent_fd, &len, sizeof (len));
         ret = write (parent_fd, value, len);
     }
 
-    ret = read (from_parent_fd, &i, sizeof (int));
+    ret = read (from_parent_fd, &i, sizeof (i));
     if (ctx)
-        ret = read (from_parent_fd, ctx, sizeof (file_op_context_t));
+        ret = read (from_parent_fd, ctx, sizeof (*ctx));
 
     (void) ret;
     return i;
@@ -467,14 +469,14 @@ parent_va_call_string (void *routine, int argc, va_list ap)
 
         len = va_arg (ap, int);
         value = va_arg (ap, void *);
-        if ((write (parent_fd, &len, sizeof (int)) != sizeof (int)) ||
+        if ((write (parent_fd, &len, sizeof (len)) != sizeof (len)) ||
             (write (parent_fd, value, len) != len))
         {
             return NULL;
         }
     }
 
-    if (read (from_parent_fd, &i, sizeof (int)) != sizeof (int))
+    if (read (from_parent_fd, &i, sizeof (i)) != sizeof (i))
         return NULL;
     if (!i)
         return NULL;
