@@ -1,7 +1,7 @@
 /*
    Editor text keep buffer.
 
-   Copyright (C) 2013-2015
+   Copyright (C) 2013-2018
    Free Software Foundation, Inc.
 
    Written by:
@@ -69,7 +69,7 @@
  * fin.
  *
  *
- * This is called a "gab buffer".
+ * This is called a "gap buffer".
  * See also:
  * http://en.wikipedia.org/wiki/Gap_buffer
  * http://stackoverflow.com/questions/4199694/data-structure-for-text-editor
@@ -330,8 +330,8 @@ edit_buffer_count_lines (const edit_buffer_t * buf, off_t first, off_t last)
 {
     long lines = 0;
 
-    first = max (first, 0);
-    last = min (last, buf->size);
+    first = MAX (first, 0);
+    last = MIN (last, buf->size);
 
     while (first < last)
         if (edit_buffer_get_byte (buf, first++) == '\n')
@@ -401,7 +401,7 @@ edit_buffer_get_word_from_pos (const edit_buffer_t * buf, off_t start_pos, off_t
                                gsize * cut)
 {
     off_t word_start;
-    long cut_len = 0;
+    gsize cut_len = 0;
     GString *match_expr;
     int c1, c2;
 
@@ -457,6 +457,9 @@ edit_buffer_insert (edit_buffer_t * buf, int c)
 
     /* update cursor position */
     buf->curs1++;
+
+    /* update file length */
+    buf->size++;
 }
 
 /* --------------------------------------------------------------------------------------------- */
@@ -486,6 +489,9 @@ edit_buffer_insert_ahead (edit_buffer_t * buf, int c)
 
     /* update cursor position */
     buf->curs2++;
+
+    /* update file length */
+    buf->size++;
 }
 
 /* --------------------------------------------------------------------------------------------- */
@@ -513,14 +519,19 @@ edit_buffer_delete (edit_buffer_t * buf)
 
     if (i == 0)
     {
-        i = buf->b2->len - 1;
-        b = g_ptr_array_index (buf->b2, i);
-        g_ptr_array_remove_index (buf->b2, i);
+        guint j;
+
+        j = buf->b2->len - 1;
+        b = g_ptr_array_index (buf->b2, j);
+        g_ptr_array_remove_index (buf->b2, j);
         g_free (b);
     }
 
     buf->curs2 = prev;
 
+    /* update file length */
+    buf->size--;
+
     return c;
 }
 
@@ -549,18 +560,22 @@ edit_buffer_backspace (edit_buffer_t * buf)
 
     if (i == 0)
     {
-        i = buf->b1->len - 1;
-        b = g_ptr_array_index (buf->b1, i);
-        g_ptr_array_remove_index (buf->b1, i);
+        guint j;
+
+        j = buf->b1->len - 1;
+        b = g_ptr_array_index (buf->b1, j);
+        g_ptr_array_remove_index (buf->b1, j);
         g_free (b);
     }
 
     buf->curs1 = prev;
 
+    /* update file length */
+    buf->size--;
+
     return c;
 }
 
-
 /* --------------------------------------------------------------------------------------------- */
 /**
  * Calculate forward offset with specified number of lines.
@@ -576,12 +591,12 @@ edit_buffer_backspace (edit_buffer_t * buf)
  */
 
 off_t
-edit_buffer_move_forward (const edit_buffer_t * buf, off_t current, long lines, off_t upto)
+edit_buffer_get_forward_offset (const edit_buffer_t * buf, off_t current, long lines, off_t upto)
 {
     if (upto != 0)
         return (off_t) edit_buffer_count_lines (buf, current, upto);
 
-    lines = max (lines, 0);
+    lines = MAX (lines, 0);
 
     while (lines-- != 0)
     {
@@ -608,9 +623,9 @@ edit_buffer_move_forward (const edit_buffer_t * buf, off_t current, long lines,
  */
 
 off_t
-edit_buffer_move_backward (const edit_buffer_t * buf, off_t current, long lines)
+edit_buffer_get_backward_offset (const edit_buffer_t * buf, off_t current, long lines)
 {
-    lines = max (lines, 0);
+    lines = MAX (lines, 0);
     current = edit_buffer_get_bol (buf, current);
 
     while (lines-- != 0 && current != 0)
