@@ -1,7 +1,7 @@
 /*
    Return a list of mounted file systems
 
-   Copyright (C) 1991-2015
+   Copyright (C) 1991-2018
    Free Software Foundation, Inc.
 
    This file is part of the Midnight Commander.
@@ -75,9 +75,6 @@
 #ifdef HAVE_SYS_FS_S5PARAM_H    /* Fujitsu UXP/V */
 #include <sys/fs/s5param.h>
 #endif
-#if defined HAVE_SYS_FILSYS_H && !defined _CRAY
-#include <sys/filsys.h>         /* SVR2 */
-#endif
 #ifdef HAVE_SYS_STATFS_H
 #include <sys/statfs.h>
 #endif
@@ -87,6 +84,7 @@
 
 #ifdef MOUNTED_GETMNTENT1       /* 4.3BSD, SunOS, HP-UX, Dynix, Irix.  */
 #include <mntent.h>
+#include <sys/types.h>
 #ifndef MOUNTED
 #ifdef _PATH_MOUNTED            /* GNU libc  */
 #define MOUNTED _PATH_MOUNTED
@@ -118,10 +116,6 @@
 #include <dirent.h>
 #endif
 
-#ifdef MOUNTED_FREAD            /* SVR2.  */
-#include <mnttab.h>
-#endif
-
 #ifdef MOUNTED_FREAD_FSTYP      /* SVR3.  */
 #include <mnttab.h>
 #include <sys/fstyp.h>
@@ -157,12 +151,6 @@
 #include <sys/mntent.h>
 #endif
 
-#ifdef MOUNTED_PROC_MOUNTINFO
-/* Use /proc/self/mountinfo instead of /proc/self/mounts (/etc/mtab)
- * on Linux, if available */
-#include <libmount/libmount.h>
-#endif
-
 #ifndef HAVE_HASMNTOPT
 #define hasmntopt(mnt, opt) ((char *) 0)
 #endif
@@ -191,6 +179,7 @@
 
 #include "lib/global.h"
 #include "lib/strutil.h"        /* str_verscmp() */
+#include "lib/unixcompat.h"     /* makedev */
 #include "mountlist.h"
 
 /*** global variables ****************************************************************************/
@@ -269,12 +258,14 @@ me_remote (char const *fs_name, char const *fs_type)
 #endif
 #ifndef ME_REMOTE
 /* A file system is 'remote' if its Fs_name contains a ':'
-   or if (it is of type (smbfs or cifs) and its Fs_name starts with '//').  */
+   or if (it is of type (smbfs or cifs) and its Fs_name starts with '//')
+   or Fs_name is equal to "-hosts" (used by autofs to mount remote fs).  */
 #define ME_REMOTE(Fs_name, Fs_type) \
     (strchr (Fs_name, ':') != NULL \
      || ((Fs_name)[0] == '/' \
          && (Fs_name)[1] == '/' \
-         && (strcmp (Fs_type, "smbfs") == 0 || strcmp (Fs_type, "cifs") == 0)))
+         && (strcmp (Fs_type, "smbfs") == 0 || strcmp (Fs_type, "cifs") == 0)) \
+     || (strcmp("-hosts", Fs_name) == 0))
 #endif
 
 /* Many space usage primitives use all 1 bits to denote a value that is
@@ -311,17 +302,6 @@ me_remote (char const *fs_name, char const *fs_type)
 #endif
 #endif
 
-#ifdef STAT_READ_FILSYS         /* SVR2 */
-/* Set errno to zero upon EOF.  */
-#define ZERO_BYTE_TRANSFER_ERRNO 0
-
-#ifdef EINTR
-#define IS_EINTR(x) ((x) == EINTR)
-#else
-#define IS_EINTR(x) 0
-#endif
-#endif /* STAT_READ_FILSYS */
-
 /*** file scope type declarations ****************************************************************/
 
 /* A mount table entry. */
@@ -329,12 +309,13 @@ struct mount_entry
 {
     char *me_devname;           /* Device node name, including "/dev/". */
     char *me_mountdir;          /* Mount point directory name. */
+    char *me_mntroot;           /* Directory on filesystem of device used
+                                   as root for the (bind) mount. */
     char *me_type;              /* "nfs", "4.2", etc. */
     dev_t me_dev;               /* Device number of me_mountdir. */
     unsigned int me_dummy:1;    /* Nonzero for dummy file systems. */
     unsigned int me_remote:1;   /* Nonzero for remote fileystems. */
     unsigned int me_type_malloced:1;    /* Nonzero if me_type was malloced. */
-    struct mount_entry *me_next;
 };
 
 struct fs_usage
@@ -351,7 +332,7 @@ struct fs_usage
 /*** file scope variables ************************************************************************/
 
 #ifdef HAVE_INFOMOUNT_LIST
-static struct mount_entry *mc_mount_list = NULL;
+static GSList *mc_mount_list = NULL;
 #endif /* HAVE_INFOMOUNT_LIST */
 
 /*** file scope functions ************************************************************************/
@@ -384,10 +365,11 @@ statvfs_works (void)
 static void
 free_mount_entry (struct mount_entry *me)
 {
-    if (!me)
+    if (me == NULL)
         return;
     g_free (me->me_devname);
     g_free (me->me_mountdir);
+    g_free (me->me_mntroot);
     if (me->me_type_malloced)
         g_free (me->me_type);
     g_free (me);
@@ -613,22 +595,47 @@ statfs (char *file, struct statfs *fsb)
 
 /* --------------------------------------------------------------------------------------------- */
 
-/* Return a list of the currently mounted file systems, or NULL on error.
-   Add each entry to the tail of the list so that they stay in order.
-   If NEED_FS_TYPE is true, ensure that the file system type fields in
-   the returned list are valid.  Otherwise, they might not be.  */
+#if defined MOUNTED_GETMNTENT1 && defined __linux__
 
-static struct mount_entry *
-read_file_system_list (int need_fs_type)
+/* Unescape the paths in mount tables.
+   STR is updated in place.  */
+static void
+unescape_tab (char *str)
 {
-    struct mount_entry *mount_list;
+    size_t i, j = 0;
+    size_t len;
+
+    len = strlen (str) + 1;
+
+    for (i = 0; i < len; i++)
+    {
+        if (str[i] == '\\' && (i + 4 < len)
+            && str[i + 1] >= '0' && str[i + 1] <= '3'
+            && str[i + 2] >= '0' && str[i + 2] <= '7' && str[i + 3] >= '0' && str[i + 3] <= '7')
+        {
+            str[j++] = (str[i + 1] - '0') * 64 + (str[i + 2] - '0') * 8 + (str[i + 3] - '0');
+            i += 3;
+        }
+        else
+            str[j++] = str[i];
+    }
+}
+#endif
+
+/* --------------------------------------------------------------------------------------------- */
+
+/* Return a list of the currently mounted file systems, or NULL on error.
+   Add each entry to the tail of the list so that they stay in order. */
+
+static GSList *
+read_file_system_list (void)
+{
+    GSList *mount_list = NULL;
     struct mount_entry *me;
-    struct mount_entry **mtail = &mount_list;
 
 #ifdef MOUNTED_LISTMNTENT
     {
         struct tabmntent *mntlist, *p;
-        struct mount_entry *me;
 
         /* the third and fourth arguments could be used to filter mounts,
            but Crays doesn't seem to have any mounts that we want to
@@ -644,13 +651,14 @@ read_file_system_list (int need_fs_type)
             me = g_malloc (sizeof (*me));
             me->me_devname = g_strdup (mnt->mnt_fsname);
             me->me_mountdir = g_strdup (mnt->mnt_dir);
+            me->me_mntroot = NULL;
             me->me_type = g_strdup (mnt->mnt_type);
             me->me_type_malloced = 1;
             me->me_dummy = ME_DUMMY (me->me_devname, me->me_type);
             me->me_remote = ME_REMOTE (me->me_devname, me->me_type);
             me->me_dev = -1;
-            *mtail = me;
-            mtail = &me->me_next;
+
+            mount_list = g_slist_prepend (mount_list, me);
         }
         freemntlist (mntlist);
     }
@@ -658,49 +666,98 @@ read_file_system_list (int need_fs_type)
 
 #ifdef MOUNTED_GETMNTENT1       /* GNU/Linux, 4.3BSD, SunOS, HP-UX, Dynix, Irix.  */
     {
-#ifdef MOUNTED_PROC_MOUNTINFO
-        struct libmnt_table *fstable = NULL;
+        FILE *fp;
 
-        fstable = mnt_new_table_from_file ("/proc/self/mountinfo");
+#ifdef __linux__
+        /* Try parsing mountinfo first, as that make device IDs available.
+           Note we could use libmount routines to simplify this parsing a little
+           (and that code is in previous versions of this function), however
+           libmount depends on libselinux which pulls in many dependencies.  */
+        char const *mountinfo = "/proc/self/mountinfo";
 
-        if (fstable != NULL)
+        fp = fopen (mountinfo, "r");
+        if (fp != NULL)
         {
-            struct libmnt_fs *fs;
-            struct libmnt_iter *iter;
+            char *line = NULL;
+            size_t buf_size = 0;
 
-            iter = mnt_new_iter (MNT_ITER_FORWARD);
-
-            while (iter && mnt_table_next_fs (fstable, iter, &fs) == 0)
+            while (getline (&line, &buf_size, fp) != -1)
             {
+                unsigned int devmaj, devmin;
+                int target_s, target_e, type_s, type_e;
+                int source_s, source_e, mntroot_s, mntroot_e;
+                char test;
+                char *dash;
+                int rc;
+
+                rc = sscanf (line, "%*u "       /* id - discarded  */
+                             "%*u "     /* parent - discarded */
+                             "%u:%u "   /* dev major:minor  */
+                             "%n%*s%n " /* mountroot */
+                             "%n%*s%n"  /* target, start and end  */
+                             "%c",      /* more data...  */
+                             &devmaj, &devmin, &mntroot_s, &mntroot_e, &target_s, &target_e, &test);
+
+                if (rc != 3 && rc != 7) /* 7 if %n included in count.  */
+                    continue;
+
+                /* skip optional fields, terminated by " - "  */
+                dash = strstr (line + target_e, " - ");
+                if (dash == NULL)
+                    continue;
+
+                rc = sscanf (dash, " - "        /* */
+                             "%n%*s%n " /* FS type, start and end  */
+                             "%n%*s%n " /* source, start and end  */
+                             "%c",      /* more data...  */
+                             &type_s, &type_e, &source_s, &source_e, &test);
+                if (rc != 1 && rc != 5) /* 5 if %n included in count.  */
+                    continue;
+
+                /* manipulate the sub-strings in place.  */
+                line[mntroot_e] = '\0';
+                line[target_e] = '\0';
+                dash[type_e] = '\0';
+                dash[source_e] = '\0';
+                unescape_tab (dash + source_s);
+                unescape_tab (line + target_s);
+                unescape_tab (line + mntroot_s);
+
                 me = g_malloc (sizeof *me);
 
-                me->me_devname = g_strdup (mnt_fs_get_source (fs));
-                me->me_mountdir = g_strdup (mnt_fs_get_target (fs));
-                me->me_type = g_strdup (mnt_fs_get_fstype (fs));
+                me->me_devname = g_strdup (dash + source_s);
+                me->me_mountdir = g_strdup (line + target_s);
+                me->me_mntroot = g_strdup (line + mntroot_s);
+                me->me_type = g_strdup (dash + type_s);
                 me->me_type_malloced = 1;
-                me->me_dev = mnt_fs_get_devno (fs);
-                /* Note we don't use mnt_fs_is_pseudofs() or mnt_fs_is_netfs() here
-                   as libmount's classification is non-compatible currently.
-                   Also we pass "false" for the "Bind" option as that's only
+                me->me_dev = makedev (devmaj, devmin);
+                /* we pass "false" for the "Bind" option as that's only
                    significant when the Fs_type is "none" which will not be
                    the case when parsing "/proc/self/mountinfo", and only
                    applies for static /etc/mtab files.  */
                 me->me_dummy = ME_DUMMY (me->me_devname, me->me_type, FALSE);
                 me->me_remote = ME_REMOTE (me->me_devname, me->me_type);
 
-                /* Add to the linked list. */
-                *mtail = me;
-                mtail = &me->me_next;
+                mount_list = g_slist_prepend (mount_list, me);
             }
 
-            mnt_free_iter (iter);
-            mnt_free_table (fstable);
+            free (line);
 
+            if (ferror (fp) != 0)
+            {
+                int saved_errno = errno;
+
+                fclose (fp);
+                errno = saved_errno;
+                goto free_then_fail;
+            }
+
+            if (fclose (fp) == EOF)
+                goto free_then_fail;
         }
-        else                    /* fallback to /proc/self/mounts (/etc/mtab) if anything failed */
-#endif /* MOUNTED_PROC_MOUNTINFO */
+        else                    /* fallback to /proc/self/mounts (/etc/mtab).  */
+#endif /* __linux __ */
         {
-            FILE *fp;
             struct mntent *mnt;
             const char *table = MOUNTED;
 
@@ -717,15 +774,14 @@ read_file_system_list (int need_fs_type)
                 me = g_malloc (sizeof (*me));
                 me->me_devname = g_strdup (mnt->mnt_fsname);
                 me->me_mountdir = g_strdup (mnt->mnt_dir);
+                me->me_mntroot = NULL;
                 me->me_type = g_strdup (mnt->mnt_type);
                 me->me_type_malloced = 1;
                 me->me_dummy = ME_DUMMY (me->me_devname, me->me_type, bind);
                 me->me_remote = ME_REMOTE (me->me_devname, me->me_type);
                 me->me_dev = dev_from_mount_options (mnt->mnt_opts);
 
-                /* Add to the linked list. */
-                *mtail = me;
-                mtail = &me->me_next;
+                mount_list = g_slist_prepend (mount_list, me);
             }
 
             if (endmntent (fp) == 0)
@@ -749,15 +805,14 @@ read_file_system_list (int need_fs_type)
             me = g_malloc (sizeof (*me));
             me->me_devname = g_strdup (fsp->f_mntfromname);
             me->me_mountdir = g_strdup (fsp->f_mntonname);
+            me->me_mntroot = NULL;
             me->me_type = fs_type;
             me->me_type_malloced = 0;
             me->me_dummy = ME_DUMMY (me->me_devname, me->me_type);
             me->me_remote = ME_REMOTE (me->me_devname, me->me_type);
             me->me_dev = (dev_t) (-1);  /* Magic; means not known yet. */
 
-            /* Add to the linked list. */
-            *mtail = me;
-            mtail = &me->me_next;
+            mount_list = g_slist_prepend (mount_list, me);
         }
     }
 #endif /* MOUNTED_GETMNTINFO */
@@ -775,15 +830,14 @@ read_file_system_list (int need_fs_type)
             me = g_malloc (sizeof (*me));
             me->me_devname = g_strdup (fsp->f_mntfromname);
             me->me_mountdir = g_strdup (fsp->f_mntonname);
+            me->me_mntroot = NULL;
             me->me_type = g_strdup (fsp->f_fstypename);
             me->me_type_malloced = 1;
             me->me_dummy = ME_DUMMY (me->me_devname, me->me_type);
             me->me_remote = ME_REMOTE (me->me_devname, me->me_type);
             me->me_dev = (dev_t) (-1);  /* Magic; means not known yet. */
 
-            /* Add to the linked list. */
-            *mtail = me;
-            mtail = &me->me_next;
+            mount_list = g_slist_prepend (mount_list, me);
         }
     }
 #endif /* MOUNTED_GETMNTINFO2 */
@@ -806,15 +860,14 @@ read_file_system_list (int need_fs_type)
             me = g_malloc (sizeof (*me));
             me->me_devname = g_strdup (fsd.fd_req.devname);
             me->me_mountdir = g_strdup (fsd.fd_req.path);
+            me->me_mntroot = NULL;
             me->me_type = gt_names[fsd.fd_req.fstype];
             me->me_type_malloced = 0;
             me->me_dummy = ME_DUMMY (me->me_devname, me->me_type);
             me->me_remote = ME_REMOTE (me->me_devname, me->me_type);
             me->me_dev = fsd.fd_req.dev;
 
-            /* Add to the linked list. */
-            *mtail = me;
-            mtail = &me->me_next;
+            mount_list = g_slist_prepend (mount_list, me);
         }
     }
 #endif /* MOUNTED_GETMNT. */
@@ -898,21 +951,20 @@ read_file_system_list (int need_fs_type)
                 me->me_devname =
                     g_strdup (fi.device_name[0] != '\0' ? fi.device_name : fi.fsh_name);
                 me->me_mountdir = g_strdup (re != NULL ? re->name : fi.fsh_name);
+                me->me_mntroot = NULL;
                 me->me_type = g_strdup (fi.fsh_name);
                 me->me_type_malloced = 1;
                 me->me_dev = fi.dev;
                 me->me_dummy = 0;
                 me->me_remote = (fi.flags & B_FS_IS_SHARED) != 0;
 
-                /* Add to the linked list. */
-                *mtail = me;
-                mtail = &me->me_next;
+                mount_list = g_slist_prepend (mount_list, me);
             }
-        *mtail = NULL;
 
         while (rootdir_list != NULL)
         {
             struct rootdir_entry *re = rootdir_list;
+
             rootdir_list = re->next;
             g_free (re->name);
             g_free (re);
@@ -950,15 +1002,14 @@ read_file_system_list (int need_fs_type)
             me = g_malloc (sizeof (*me));
             me->me_devname = g_strdup (stats[counter].f_mntfromname);
             me->me_mountdir = g_strdup (stats[counter].f_mntonname);
+            me->me_mntroot = NULL;
             me->me_type = g_strdup (FS_TYPE (stats[counter]));
             me->me_type_malloced = 1;
             me->me_dummy = ME_DUMMY (me->me_devname, me->me_type);
             me->me_remote = ME_REMOTE (me->me_devname, me->me_type);
             me->me_dev = (dev_t) (-1);  /* Magic; means not known yet. */
 
-            /* Add to the linked list. */
-            *mtail = me;
-            mtail = &me->me_next;
+            mount_list = g_slist_prepend (mount_list, me);
         }
 
         g_free (stats);
@@ -984,11 +1035,11 @@ read_file_system_list (int need_fs_type)
             me->me_devname = g_strconcat ("/dev/", mnt.mt_dev, (char *) NULL);
 #endif
             me->me_mountdir = g_strdup (mnt.mt_filsys);
+            me->me_mntroot = NULL;
             me->me_dev = (dev_t) (-1);  /* Magic; means not known yet. */
             me->me_type = "";
             me->me_type_malloced = 0;
 #ifdef GETFSTYP                 /* SVR3.  */
-            if (need_fs_type)
             {
                 struct statfs fsd;
                 char typebuf[FSTYPSZ];
@@ -1004,15 +1055,14 @@ read_file_system_list (int need_fs_type)
             me->me_dummy = ME_DUMMY (me->me_devname, me->me_type);
             me->me_remote = ME_REMOTE (me->me_devname, me->me_type);
 
-            /* Add to the linked list. */
-            *mtail = me;
-            mtail = &me->me_next;
+            mount_list = g_slist_prepend (mount_list, me);
         }
 
         if (ferror (fp))
         {
             /* The last fread() call must have failed.  */
             int saved_errno = errno;
+
             fclose (fp);
             errno = saved_errno;
             goto free_then_fail;
@@ -1026,20 +1076,20 @@ read_file_system_list (int need_fs_type)
 #ifdef MOUNTED_GETMNTTBL        /* DolphinOS goes its own way.  */
     {
         struct mntent **mnttbl = getmnttbl (), **ent;
+
         for (ent = mnttbl; *ent; ent++)
         {
             me = g_malloc (sizeof (*me));
             me->me_devname = g_strdup ((*ent)->mt_resource);
             me->me_mountdir = g_strdup ((*ent)->mt_directory);
+            me->me_mntroot = NULL;
             me->me_type = g_strdup ((*ent)->mt_fstype);
             me->me_type_malloced = 1;
             me->me_dummy = ME_DUMMY (me->me_devname, me->me_type);
             me->me_remote = ME_REMOTE (me->me_devname, me->me_type);
             me->me_dev = (dev_t) (-1);  /* Magic; means not known yet. */
 
-            /* Add to the linked list. */
-            *mtail = me;
-            mtail = &me->me_next;
+            mount_list = g_slist_prepend (mount_list, me);
         }
         endmnttbl ();
     }
@@ -1094,15 +1144,14 @@ read_file_system_list (int need_fs_type)
                 me = g_malloc (sizeof (*me));
                 me->me_devname = g_strdup (mnt.mnt_special);
                 me->me_mountdir = g_strdup (mnt.mnt_mountp);
+                me->me_mntroot = NULL;
                 me->me_type = g_strdup (mnt.mnt_fstype);
                 me->me_type_malloced = 1;
                 me->me_dummy = MNT_IGNORE (&mnt) != 0;
                 me->me_remote = ME_REMOTE (me->me_devname, me->me_type);
                 me->me_dev = dev_from_mount_options (mnt.mnt_mntopts);
 
-                /* Add to the linked list. */
-                *mtail = me;
-                mtail = &me->me_next;
+                mount_list = g_slist_prepend (mount_list, me);
             }
 
             ret = fclose (fp) == EOF ? errno : 0 < ret ? 0 : -1;
@@ -1122,21 +1171,24 @@ read_file_system_list (int need_fs_type)
 #ifdef MOUNTED_VMOUNT           /* AIX.  */
     {
         int bufsize;
-        char *entries, *thisent;
+        void *entries;
+        char *thisent;
         struct vmount *vmp;
         int n_entries;
         int i;
 
         /* Ask how many bytes to allocate for the mounted file system info.  */
-        if (mntctl (MCTL_QUERY, sizeof (bufsize), (struct vmount *) &bufsize) != 0)
+        entries = &bufsize;
+        if (mntctl (MCTL_QUERY, sizeof (bufsize), entries) != 0)
             return NULL;
         entries = g_malloc (bufsize);
 
         /* Get the list of mounted file systems.  */
-        n_entries = mntctl (MCTL_QUERY, bufsize, (struct vmount *) entries);
+        n_entries = mntctl (MCTL_QUERY, bufsize, entries);
         if (n_entries < 0)
         {
             int saved_errno = errno;
+
             g_free (entries);
             errno = saved_errno;
             return NULL;
@@ -1164,6 +1216,7 @@ read_file_system_list (int need_fs_type)
                 me->me_devname = g_strdup (thisent + vmp->vmt_data[VMT_OBJECT].vmt_off);
             }
             me->me_mountdir = g_strdup (thisent + vmp->vmt_data[VMT_STUB].vmt_off);
+            me->me_mntroot = NULL;
             me->me_type = g_strdup (fstype_to_string (vmp->vmt_gfstype));
             me->me_type_malloced = 1;
             options = thisent + vmp->vmt_data[VMT_ARGS].vmt_off;
@@ -1174,15 +1227,12 @@ read_file_system_list (int need_fs_type)
                                 || ignore[sizeof ("ignore") - 1] == '\0'));
             me->me_dev = (dev_t) (-1);  /* vmt_fsid might be the info we want.  */
 
-            /* Add to the linked list. */
-            *mtail = me;
-            mtail = &me->me_next;
+            mount_list = g_slist_prepend (mount_list, me);
         }
         g_free (entries);
     }
 #endif /* MOUNTED_VMOUNT. */
 
-
 #ifdef MOUNTED_INTERIX_STATVFS
     {
         DIR *dirp = opendir ("/dev/fs");
@@ -1208,47 +1258,32 @@ read_file_system_list (int need_fs_type)
                 me = g_malloc (sizeof *me);
                 me->me_devname = g_strdup (dev.f_mntfromname);
                 me->me_mountdir = g_strdup (dev.f_mntonname);
+                me->me_mntroot = NULL;
                 me->me_type = g_strdup (dev.f_fstypename);
                 me->me_type_malloced = 1;
                 me->me_dummy = ME_DUMMY (me->me_devname, me->me_type);
                 me->me_remote = ME_REMOTE (me->me_devname, me->me_type);
                 me->me_dev = (dev_t) (-1);      /* Magic; means not known yet. */
 
-                /* Add to the linked list. */
-                *mtail = me;
-                mtail = &me->me_next;
+                mount_list = g_slist_prepend (mount_list, me);
             }
         }
         closedir (dirp);
     }
 #endif /* MOUNTED_INTERIX_STATVFS */
 
-    (void) need_fs_type;        /* avoid argument-unused warning */
-    *mtail = NULL;
-    return mount_list;
-
+    return g_slist_reverse (mount_list);
 
   free_then_fail:
     {
         int saved_errno = errno;
-        *mtail = NULL;
 
-        while (mount_list)
-        {
-            me = mount_list->me_next;
-            g_free (mount_list->me_devname);
-            g_free (mount_list->me_mountdir);
-            if (mount_list->me_type_malloced)
-                g_free (mount_list->me_type);
-            g_free (mount_list);
-            mount_list = me;
-        }
+        g_slist_free_full (mount_list, (GDestroyNotify) free_mount_entry);
 
         errno = saved_errno;
         return NULL;
     }
 }
-
 #endif /* HAVE_INFOMOUNT_LIST */
 
 /* --------------------------------------------------------------------------------------------- */
@@ -1258,28 +1293,33 @@ read_file_system_list (int need_fs_type)
  ** QNX has no [gs]etmnt*(), [gs]etfs*(), or /etc/mnttab, but can do
  ** this via the following code.
  ** Note that, as this is based on CWD, it only fills one mount_entry
- ** structure. See my_statfs() in utilunix.c for the "other side" of
- ** this hack.
+ ** structure. See my_statfs() below for the "other side" of this hack.
  */
 
-static struct mount_entry *
-read_file_system_list (int need_fs_type, int all_fs)
+static GSList *
+read_file_system_list (void)
 {
     struct _disk_entry de;
     struct statfs fs;
     int i, fd;
     char *tp, dev[_POSIX_NAME_MAX], dir[_POSIX_PATH_MAX];
+    struct mount_entry *me = NULL;
+    static GSList *list = NULL;
 
-    static struct mount_entry *me = NULL;
-
-    if (me)
+    if (list != NULL)
     {
+        me = (struct mount_entry *) list->data;
+
         g_free (me->me_devname);
         g_free (me->me_mountdir);
+        g_free (me->me_mntroot);
         g_free (me->me_type);
     }
     else
+    {
         me = (struct mount_entry *) g_malloc (sizeof (struct mount_entry));
+        list = g_slist_prepend (list, me);
+    }
 
     if (!getcwd (dir, _POSIX_PATH_MAX))
         return (NULL);
@@ -1330,6 +1370,7 @@ read_file_system_list (int need_fs_type, int all_fs)
 
     me->me_devname = g_strdup (dev);
     me->me_mountdir = g_strdup (dir);
+    me->me_mntroot = NULL;
     me->me_type = g_strdup (tp);
     me->me_dev = de.disk_type;
 
@@ -1341,82 +1382,12 @@ read_file_system_list (int need_fs_type, int all_fs)
     fprintf (stderr, "fsys_get_mount_pt():\n\tmount point='%s'\n", dir);
 #endif /* DEBUG */
 
-    return (me);
+    return (list);
 }
 #endif /* HAVE_INFOMOUNT_QNX */
 
 /* --------------------------------------------------------------------------------------------- */
 
-#ifdef STAT_READ_FILSYS         /* SVR2 */
-
-/* Read(write) up to COUNT bytes at BUF from(to) descriptor FD, retrying if
-   interrupted.  Return the actual number of bytes read(written), zero for EOF,
-   or SAFE_READ_ERROR(SAFE_WRITE_ERROR) upon error.  */
-static size_t
-safe_read (int fd, void *buf, size_t count)
-{
-    /* Work around a bug in Tru64 5.1.  Attempting to read more than
-       INT_MAX bytes fails with errno == EINVAL.  See
-       <http://lists.gnu.org/archive/html/bug-gnu-utils/2002-04/msg00010.html>.
-       When decreasing COUNT, keep it block-aligned.  */
-    /* *INDENT-OFF* */
-    enum { BUGGY_READ_MAXIMUM = INT_MAX & ~8191 };
-    /* *INDENT-ON* */
-
-    while (TRUE)
-    {
-        ssize_t result;
-
-        result = read (fd, buf, count);
-
-        if (0 <= result)
-            return result;
-        else if (IS_EINTR (errno))
-            continue;
-        else if (errno == EINVAL && BUGGY_READ_MAXIMUM < count)
-            count = BUGGY_READ_MAXIMUM;
-        else
-            return result;
-    }
-}
-
-/* --------------------------------------------------------------------------------------------- */
-
-/* Read COUNT bytes at BUF to(from) descriptor FD, retrying if
-   interrupted or if a partial write(read) occurs.  Return the number
-   of bytes transferred.
-   When writing, set errno if fewer than COUNT bytes are written.
-   When reading, if fewer than COUNT bytes are read, you must examine
-   errno to distinguish failure from EOF (errno == 0).  */
-
-static size_t
-full_read (int fd, void *buf, size_t count)
-{
-    size_t total = 0;
-    char *ptr = (char *) buf;
-
-    while (count > 0)
-    {
-        size_t n_rw = safe_read (fd, ptr, count);
-        if (n_rw == (size_t) (-1))
-            break;
-        if (n_rw == 0)
-        {
-            errno = ZERO_BYTE_TRANSFER_ERRNO;
-            break;
-        }
-        total += n_rw;
-        ptr += n_rw;
-        count -= n_rw;
-    }
-
-    return total;
-}
-
-#endif /* STAT_READ_FILSYS */
-
-/* --------------------------------------------------------------------------------------------- */
-
 #ifdef HAVE_INFOMOUNT
 /* Fill in the fields of FSP with information about space usage for
    the file system on which FILE resides.
@@ -1482,40 +1453,6 @@ get_fs_usage (char const *file, char const *disk, struct fs_usage *fsp)
         fsp->fsu_files = PROPAGATE_ALL_ONES (fsd.fd_req.gtot);
         fsp->fsu_ffree = PROPAGATE_ALL_ONES (fsd.fd_req.gfree);
 
-#elif defined STAT_READ_FILSYS  /* SVR2 */
-#ifndef SUPERBOFF
-#define SUPERBOFF (SUPERB * 512)
-#endif
-
-        struct filsys fsd;
-        int fd;
-
-        if (!disk)
-        {
-            errno = 0;
-            return -1;
-        }
-
-        fd = open (disk, O_RDONLY);
-        if (fd < 0)
-            return -1;
-        lseek (fd, (off_t) SUPERBOFF, 0);
-        if (full_read (fd, (char *) &fsd, sizeof (fsd)) != sizeof (fsd))
-        {
-            close (fd);
-            return -1;
-        }
-        close (fd);
-
-        fsp->fsu_blocksize = (fsd.s_type == Fs2b ? 1024 : 512);
-        fsp->fsu_blocks = PROPAGATE_ALL_ONES (fsd.s_fsize);
-        fsp->fsu_bfree = PROPAGATE_ALL_ONES (fsd.s_tfree);
-        fsp->fsu_bavail = PROPAGATE_TOP_BIT (fsd.s_tfree);
-        fsp->fsu_bavail_top_bit_set = EXTRACT_TOP_BIT (fsd.s_tfree) != 0;
-        fsp->fsu_files = (fsd.s_isize == -1
-                          ? UINTMAX_MAX : (fsd.s_isize - 2) * INOPB * (fsd.s_type == Fs2b ? 2 : 1));
-        fsp->fsu_ffree = PROPAGATE_ALL_ONES (fsd.s_tinode);
-
 #elif defined STAT_STATFS3_OSF1 /* OSF/1 */
 
         struct statfs fsd;
@@ -1620,15 +1557,7 @@ void
 free_my_statfs (void)
 {
 #ifdef HAVE_INFOMOUNT_LIST
-    while (mc_mount_list != NULL)
-    {
-        struct mount_entry *next;
-
-        next = mc_mount_list->me_next;
-        free_mount_entry (mc_mount_list);
-        mc_mount_list = next;
-    }
-
+    g_slist_free_full (mc_mount_list, (GDestroyNotify) free_mount_entry);
     mc_mount_list = NULL;
 #endif /* HAVE_INFOMOUNT_LIST */
 }
@@ -1640,7 +1569,7 @@ init_my_statfs (void)
 {
 #ifdef HAVE_INFOMOUNT_LIST
     free_my_statfs ();
-    mc_mount_list = read_file_system_list (1);
+    mc_mount_list = read_file_system_list ();
 #endif /* HAVE_INFOMOUNT_LIST */
 }
 
@@ -1652,31 +1581,33 @@ my_statfs (struct my_statfs *myfs_stats, const char *path)
 #ifdef HAVE_INFOMOUNT_LIST
     size_t len = 0;
     struct mount_entry *entry = NULL;
-    struct mount_entry *temp = mc_mount_list;
+    GSList *temp;
     struct fs_usage fs_use;
 
-    while (temp)
+    for (temp = mc_mount_list; temp != NULL; temp = g_slist_next (temp))
     {
+        struct mount_entry *me;
         size_t i;
 
-        i = strlen (temp->me_mountdir);
-        if (i > len && (strncmp (path, temp->me_mountdir, i) == 0))
-            if (entry == NULL || IS_PATH_SEP (path[i]) || path[i] == '\0')
-            {
-                len = i;
-                entry = temp;
-            }
-        temp = temp->me_next;
+        me = (struct mount_entry *) temp->data;
+        i = strlen (me->me_mountdir);
+        if (i > len && (strncmp (path, me->me_mountdir, i) == 0) &&
+            (entry == NULL || IS_PATH_SEP (path[i]) || path[i] == '\0'))
+        {
+            len = i;
+            entry = me;
+        }
     }
 
-    if (entry)
+    if (entry != NULL)
     {
-        memset (&fs_use, 0, sizeof (struct fs_usage));
+        memset (&fs_use, 0, sizeof (fs_use));
         get_fs_usage (entry->me_mountdir, NULL, &fs_use);
 
         myfs_stats->type = entry->me_dev;
         myfs_stats->typename = entry->me_type;
         myfs_stats->mpoint = entry->me_mountdir;
+        myfs_stats->mroot = entry->me_mntroot;
         myfs_stats->device = entry->me_devname;
         myfs_stats->avail =
             ((uintmax_t) (getuid ()? fs_use.fsu_bavail : fs_use.fsu_bfree) *
@@ -1690,15 +1621,14 @@ my_statfs (struct my_statfs *myfs_stats, const char *path)
 
 #ifdef HAVE_INFOMOUNT_QNX
         /*
-         ** This is the "other side" of the hack to read_file_system_list() in
-         ** mountlist.c.
+         ** This is the "other side" of the hack to read_file_system_list() above.
          ** It's not the most efficient approach, but consumes less memory. It
          ** also accommodates QNX's ability to mount filesystems on the fly.
          */
         struct mount_entry *entry;
     struct fs_usage fs_use;
 
-    entry = read_file_system_list (0, 0);
+    entry = read_file_system_list ();
     if (entry != NULL)
     {
         get_fs_usage (entry->me_mountdir, NULL, &fs_use);
@@ -1706,6 +1636,7 @@ my_statfs (struct my_statfs *myfs_stats, const char *path)
         myfs_stats->type = entry->me_dev;
         myfs_stats->typename = entry->me_type;
         myfs_stats->mpoint = entry->me_mountdir;
+        myfs_stats->mroot = entry->me_mntroot;
         myfs_stats->device = entry->me_devname;
 
         myfs_stats->avail = ((uintmax_t) fs_use.fsu_bfree * fs_use.fsu_blocksize) >> 10;
