@@ -1,7 +1,7 @@
 /*
    Virtual File System: External file system.
 
-   Copyright (C) 1995-2015
+   Copyright (C) 1995-2018
    Free Software Foundation, Inc.
 
    Written by:
@@ -246,7 +246,7 @@ extfs_find_entry_int (struct entry *dir, const char *name, GSList * list,
     if (g_path_is_absolute (name))
     {
         /* Handle absolute paths */
-        name = (char *) g_path_skip_root (name);
+        name = g_path_skip_root (name);
         dir = dir->inode->archive->root_entry;
     }
 
@@ -255,7 +255,7 @@ extfs_find_entry_int (struct entry *dir, const char *name, GSList * list,
     name_end = name + strlen (name);
 
     q = strchr (p, PATH_SEP);
-    if (q == '\0')
+    if (q == NULL)
         q = strchr (p, '\0');
 
     while ((pent != NULL) && (c != '\0') && (*p != '\0'))
@@ -308,9 +308,11 @@ extfs_find_entry_int (struct entry *dir, const char *name, GSList * list,
         }
         /* Next iteration */
         *q = c;
+        if (c == '\0')
+            break;
         p = q + 1;
         q = strchr (p, PATH_SEP);
-        if (q == '\0')
+        if (q == NULL)
             q = strchr (p, '\0');
     }
     if (pent == NULL)
@@ -848,7 +850,7 @@ extfs_cmd (const char *str_extfs_cmd, struct archive *archive,
     g_free (quoted_archive_name);
 
     open_error_pipe ();
-    retval = my_system (EXECUTE_AS_SHELL, mc_global.tty.shell, cmd);
+    retval = my_system (EXECUTE_AS_SHELL, mc_global.shell->path, cmd);
     g_free (cmd);
     close_error_pipe (D_ERROR, NULL);
     return retval;
@@ -1093,12 +1095,13 @@ extfs_stat_move (struct stat *buf, const struct inode *inode)
 #ifdef HAVE_STRUCT_STAT_ST_BLKSIZE
     buf->st_blksize = RECORDSIZE;
 #endif
-#ifdef HAVE_STRUCT_STAT_ST_BLOCKS
-    buf->st_blocks = (inode->size + RECORDSIZE - 1) / RECORDSIZE;
-#endif
+    vfs_adjust_stat (buf);
     buf->st_atime = inode->atime;
     buf->st_mtime = inode->mtime;
     buf->st_ctime = inode->ctime;
+#ifdef HAVE_STRUCT_STAT_ST_MTIM
+    buf->st_atim.tv_nsec = buf->st_mtim.tv_nsec = buf->st_ctim.tv_nsec = 0;
+#endif
 }
 
 /* --------------------------------------------------------------------------------------------- */
