@@ -1,7 +1,7 @@
 /*
    Virtual File System: GNU Tar file system.
 
-   Copyright (C) 1995-2015
+   Copyright (C) 1995-2018
    Free Software Foundation, Inc.
 
    Written by:
@@ -39,7 +39,6 @@
 #include <sys/types.h>
 #include <errno.h>
 #include <ctype.h>
-#include <fcntl.h>
 
 #ifdef hpux
 /* major() and minor() macros (among other things) defined here for hpux */
@@ -48,6 +47,7 @@
 
 #include "lib/global.h"
 #include "lib/util.h"
+#include "lib/unixcompat.h"     /* makedev() */
 #include "lib/widget.h"         /* message() */
 
 #include "lib/vfs/vfs.h"
@@ -237,7 +237,7 @@ static union record rec_buf;
  * Result is -1 if the field is invalid (all blank, or nonoctal).
  */
 static long
-tar_from_oct (int digs, char *where)
+tar_from_oct (int digs, const char *where)
 {
     long value;
 
@@ -304,8 +304,9 @@ tar_open_archive_int (struct vfs_class *me, const vfs_path_t * vpath, struct vfs
 
     /* Find out the method to handle this tar file */
     type = get_compression_type (result, archive->name);
-    mc_lseek (result, 0, SEEK_SET);
-    if (type != COMPRESSION_NONE)
+    if (type == COMPRESSION_NONE)
+        mc_lseek (result, 0, SEEK_SET);
+    else
     {
         char *s;
         vfs_path_t *tmp_vpath;
@@ -355,10 +356,10 @@ tar_get_next_record (struct vfs_s_super *archive, int tard)
 
     (void) archive;
 
-    n = mc_read (tard, rec_buf.charptr, RECORDSIZE);
-    if (n != RECORDSIZE)
+    n = mc_read (tard, rec_buf.charptr, sizeof (rec_buf.charptr));
+    if (n != sizeof (rec_buf.charptr))
         return NULL;            /* An error has occurred */
-    current_tar_position += RECORDSIZE;
+    current_tar_position += sizeof (rec_buf.charptr);
     return &rec_buf;
 }
 
@@ -369,8 +370,54 @@ tar_skip_n_records (struct vfs_s_super *archive, int tard, size_t n)
 {
     (void) archive;
 
-    mc_lseek (tard, n * RECORDSIZE, SEEK_CUR);
-    current_tar_position += n * RECORDSIZE;
+    mc_lseek (tard, n * sizeof (rec_buf.charptr), SEEK_CUR);
+    current_tar_position += n * sizeof (rec_buf.charptr);
+}
+
+/* --------------------------------------------------------------------------------------------- */
+
+static ReadStatus
+tar_checksum (const union record *header)
+{
+    long recsum;
+    long signed_sum = 0;
+    long sum = 0;
+    int i;
+    const char *p = header->charptr;
+
+    recsum = tar_from_oct (8, header->header.chksum);
+
+    for (i = sizeof (*header); --i >= 0;)
+    {
+        /*
+         * We can't use unsigned char here because of old compilers,
+         * e.g. V7.
+         */
+        signed_sum += *p;
+        sum += 0xFF & *p++;
+    }
+
+    /* Adjust checksum to count the "chksum" field as blanks. */
+    for (i = sizeof (header->header.chksum); --i >= 0;)
+    {
+        sum -= 0xFF & header->header.chksum[i];
+        signed_sum -= (char) header->header.chksum[i];
+    }
+
+    sum += ' ' * sizeof (header->header.chksum);
+    signed_sum += ' ' * sizeof (header->header.chksum);
+
+    /*
+     * This is a zeroed record...whole record is 0's except
+     * for the 8 blanks we faked for the checksum field.
+     */
+    if (sum == 8 * ' ')
+        return STATUS_EOFMARK;
+
+    if (sum != recsum && signed_sum != recsum)
+        return STATUS_BADCHECKSUM;
+
+    return STATUS_SUCCESS;
 }
 
 /* --------------------------------------------------------------------------------------------- */
@@ -387,7 +434,7 @@ tar_fill_stat (struct vfs_s_super *archive, struct stat *st, union record *heade
      * know about the other modes but I think I cause no new
      * problem when I adjust them, too. -- Norbert.
      */
-    if (header->header.linkflag == LF_DIR)
+    if (header->header.linkflag == LF_DIR || header->header.linkflag == LF_DUMPDIR)
         st->st_mode |= S_IFDIR;
     else if (header->header.linkflag == LF_SYMLINK)
         st->st_mode |= S_IFLNK;
@@ -401,7 +448,10 @@ tar_fill_stat (struct vfs_s_super *archive, struct stat *st, union record *heade
         st->st_mode |= S_IFREG;
 
     st->st_dev = 0;
+#ifdef HAVE_STRUCT_STAT_ST_RDEV
     st->st_rdev = 0;
+#endif
+
     switch (arch->type)
     {
     case TAR_USTAR:
@@ -420,15 +470,27 @@ tar_fill_stat (struct vfs_s_super *archive, struct stat *st, union record *heade
         {
         case LF_BLK:
         case LF_CHR:
+#ifdef HAVE_STRUCT_STAT_ST_RDEV
             st->st_rdev =
-                (tar_from_oct (8, header->header.devmajor) << 8) |
-                tar_from_oct (8, header->header.devminor);
+                makedev (tar_from_oct (8, header->header.devmajor),
+                         tar_from_oct (8, header->header.devminor));
+#endif
+            break;
+        default:
+            break;
         }
+        break;
+
     default:
         st->st_uid = tar_from_oct (8, header->header.uid);
         st->st_gid = tar_from_oct (8, header->header.gid);
+        break;
     }
+
     st->st_size = h_size;
+#ifdef HAVE_STRUCT_STAT_ST_MTIM
+    st->st_atim.tv_nsec = st->st_mtim.tv_nsec = st->st_ctim.tv_nsec = 0;
+#endif
     st->st_mtime = tar_from_oct (1 + 12, header->header.mtime);
     st->st_atime = 0;
     st->st_ctime = 0;
@@ -437,6 +499,11 @@ tar_fill_stat (struct vfs_s_super *archive, struct stat *st, union record *heade
         st->st_atime = tar_from_oct (1 + 12, header->header.unused.oldgnu.atime);
         st->st_ctime = tar_from_oct (1 + 12, header->header.unused.oldgnu.ctime);
     }
+
+#ifdef HAVE_STRUCT_STAT_ST_BLKSIZE
+    st->st_blksize = 8 * 1024;  /* FIXME */
+#endif
+    vfs_adjust_stat (st);
 }
 
 /* --------------------------------------------------------------------------------------------- */
@@ -449,10 +516,7 @@ static ReadStatus
 tar_read_header (struct vfs_class *me, struct vfs_s_super *archive, int tard, size_t * h_size)
 {
     tar_super_data_t *arch = (tar_super_data_t *) archive->data;
-
-    int i;
-    long sum, signed_sum, recsum;
-    char *p;
+    ReadStatus checksum_status;
     union record *header;
     static char *next_long_name = NULL, *next_long_link = NULL;
 
@@ -462,39 +526,9 @@ tar_read_header (struct vfs_class *me, struct vfs_s_super *archive, int tard, si
     if (NULL == header)
         return STATUS_EOF;
 
-    recsum = tar_from_oct (8, header->header.chksum);
-
-    sum = 0;
-    signed_sum = 0;
-    p = header->charptr;
-    for (i = sizeof (*header); --i >= 0;)
-    {
-        /*
-         * We can't use unsigned char here because of old compilers,
-         * e.g. V7.
-         */
-        signed_sum += *p;
-        sum += 0xFF & *p++;
-    }
-
-    /* Adjust checksum to count the "chksum" field as blanks. */
-    for (i = sizeof (header->header.chksum); --i >= 0;)
-    {
-        sum -= 0xFF & header->header.chksum[i];
-        signed_sum -= (char) header->header.chksum[i];
-    }
-    sum += ' ' * sizeof header->header.chksum;
-    signed_sum += ' ' * sizeof header->header.chksum;
-
-    /*
-     * This is a zeroed record...whole record is 0's except
-     * for the 8 blanks we faked for the checksum field.
-     */
-    if (sum == 8 * ' ')
-        return STATUS_EOFMARK;
-
-    if (sum != recsum && signed_sum != recsum)
-        return STATUS_BADCHECKSUM;
+    checksum_status = tar_checksum (header);
+    if (checksum_status != STATUS_SUCCESS)
+        return checksum_status;
 
     /*
      * Try to determine the archive format.
@@ -536,15 +570,10 @@ tar_read_header (struct vfs_class *me, struct vfs_s_super *archive, int tard, si
     else
         *h_size = tar_from_oct (1 + 12, header->header.size);
 
-    /*
-     * Skip over directory snapshot info records that
-     * are stored in incremental tar archives.
-     */
     if (header->header.linkflag == LF_DUMPDIR)
     {
         if (arch->type == TAR_UNKNOWN)
             arch->type = TAR_GNU;
-        return STATUS_SUCCESS;
     }
 
     /*
@@ -611,7 +640,7 @@ tar_read_header (struct vfs_class *me, struct vfs_s_super *archive, int tard, si
         struct vfs_s_entry *entry;
         struct vfs_s_inode *inode = NULL, *parent;
         off_t data_position;
-        char *q;
+        char *p, *q;
         size_t len;
         char *current_file_name, *current_link_name;
 
@@ -763,7 +792,7 @@ tar_open_archive (struct vfs_s_super *archive, const vfs_path_t * vpath,
 
     while (TRUE)
     {
-        size_t h_size;
+        size_t h_size = 0;
         ReadStatus prev_status = status;
 
         status = tar_read_header (vpath_element->class, archive, tard, &h_size);
@@ -788,7 +817,7 @@ tar_open_archive (struct vfs_s_super *archive, const vfs_path_t * vpath,
                 {
                     message (D_ERROR, MSG_ERROR, _("%s\ndoesn't look like a tar archive."),
                              vfs_path_as_str (vpath));
-                    /* FALL THRU */
+                    MC_FALLTHROUGH;
 
                     /* Error after header rec */
                 }
@@ -800,14 +829,19 @@ tar_open_archive (struct vfs_s_super *archive, const vfs_path_t * vpath,
 
             case STATUS_EOF:
                 return 0;
+
+            default:
+                break;
             }
 
             /* Record of zeroes */
         case STATUS_EOFMARK:   /* If error after 0's */
-            /* FALL THRU */
+            MC_FALLTHROUGH;
             /* exit from loop */
         case STATUS_EOF:       /* End of archive */
             break;
+        default:
+            break;
         }
         break;
     }
