@@ -2,7 +2,7 @@
    Internal file viewer for the Midnight Commander
    Function for whow info on display
 
-   Copyright (C) 1994-2015
+   Copyright (C) 1994-2018
    Free Software Foundation, Inc.
 
    Written by:
@@ -76,15 +76,15 @@ static enum ruler_type
 /** Define labels and handlers for functional keys */
 
 static void
-mcview_set_buttonbar (mcview_t * view)
+mcview_set_buttonbar (WView * view)
 {
     WDialog *h = WIDGET (view)->owner;
     WButtonBar *b = find_buttonbar (h);
-    const global_keymap_t *keymap = view->hex_mode ? viewer_hex_map : viewer_map;
+    const global_keymap_t *keymap = view->mode_flags.hex ? viewer_hex_map : viewer_map;
 
     buttonbar_set_label (b, 1, Q_ ("ButtonBar|Help"), keymap, WIDGET (view));
 
-    if (view->hex_mode)
+    if (view->mode_flags.hex)
     {
         if (view->hexedit_mode)
             buttonbar_set_label (b, 2, Q_ ("ButtonBar|View"), keymap, WIDGET (view));
@@ -100,7 +100,7 @@ mcview_set_buttonbar (mcview_t * view)
     }
     else
     {
-        buttonbar_set_label (b, 2, view->text_wrap_mode ? Q_ ("ButtonBar|UnWrap")
+        buttonbar_set_label (b, 2, view->mode_flags.wrap ? Q_ ("ButtonBar|UnWrap")
                              : Q_ ("ButtonBar|Wrap"), keymap, WIDGET (view));
         buttonbar_set_label (b, 4, Q_ ("ButtonBar|Hex"), keymap, WIDGET (view));
         buttonbar_set_label (b, 6, "", keymap, WIDGET (view));
@@ -108,16 +108,13 @@ mcview_set_buttonbar (mcview_t * view)
     }
 
     buttonbar_set_label (b, 5, Q_ ("ButtonBar|Goto"), keymap, WIDGET (view));
-    buttonbar_set_label (b, 8, view->magic_mode ? Q_ ("ButtonBar|Raw")
+    buttonbar_set_label (b, 8, view->mode_flags.magic ? Q_ ("ButtonBar|Raw")
                          : Q_ ("ButtonBar|Parse"), keymap, WIDGET (view));
 
-    if (mcview_is_in_panel (view))
-        buttonbar_set_label (b, 10, "", keymap, WIDGET (view));
-    else
+    if (!mcview_is_in_panel (view))     /* don't override some panel buttonbar keys  */
     {
-        /* don't override some panel buttonbar keys  */
         buttonbar_set_label (b, 3, Q_ ("ButtonBar|Quit"), keymap, WIDGET (view));
-        buttonbar_set_label (b, 9, view->text_nroff_mode ? Q_ ("ButtonBar|Unform")
+        buttonbar_set_label (b, 9, view->mode_flags.nroff ? Q_ ("ButtonBar|Unform")
                              : Q_ ("ButtonBar|Format"), keymap, WIDGET (view));
         buttonbar_set_label (b, 10, Q_ ("ButtonBar|Quit"), keymap, WIDGET (view));
     }
@@ -126,7 +123,7 @@ mcview_set_buttonbar (mcview_t * view)
 /* --------------------------------------------------------------------------------------------- */
 
 static void
-mcview_display_percent (mcview_t * view, off_t p)
+mcview_display_percent (WView * view, off_t p)
 {
     int percent;
 
@@ -146,7 +143,7 @@ mcview_display_percent (mcview_t * view, off_t p)
 /* --------------------------------------------------------------------------------------------- */
 
 static void
-mcview_display_status (mcview_t * view)
+mcview_display_status (WView * view)
 {
     const screen_dimen top = view->status_area.top;
     const screen_dimen left = view->status_area.left;
@@ -168,7 +165,7 @@ mcview_display_status (mcview_t * view)
     if (width > 40)
     {
         widget_move (view, top, width - 32);
-        if (view->hex_mode)
+        if (view->mode_flags.hex)
             tty_printf ("0x%08" PRIxMAX, (uintmax_t) view->hex_cursor);
         else
         {
@@ -191,7 +188,7 @@ mcview_display_status (mcview_t * view)
     else
         tty_print_string (str_fit_to_term (file_label, width - 5, J_LEFT_FIT));
     if (width > 26)
-        mcview_display_percent (view, view->hex_mode ? view->hex_cursor : view->dpy_end);
+        mcview_display_percent (view, view->mode_flags.hex ? view->hex_cursor : view->dpy_end);
 }
 
 /* --------------------------------------------------------------------------------------------- */
@@ -199,7 +196,7 @@ mcview_display_status (mcview_t * view)
 /* --------------------------------------------------------------------------------------------- */
 
 void
-mcview_update (mcview_t * view)
+mcview_update (WView * view)
 {
     static int dirt_limit = 1;
 
@@ -245,23 +242,19 @@ mcview_update (mcview_t * view)
 /** Displays as much data from view->dpy_start as fits on the screen */
 
 void
-mcview_display (mcview_t * view)
+mcview_display (WView * view)
 {
-    if (view->hex_mode)
-    {
+    if (view->mode_flags.hex)
         mcview_display_hex (view);
-    }
     else
-    {
         mcview_display_text (view);
-    }
     mcview_display_status (view);
 }
 
 /* --------------------------------------------------------------------------------------------- */
 
 void
-mcview_compute_areas (mcview_t * view)
+mcview_compute_areas (WView * view)
 {
     struct area view_area;
     screen_dimen height, rest, y;
@@ -284,12 +277,12 @@ mcview_compute_areas (mcview_t * view)
     /* Compute the heights of the areas */
     rest = view_area.height;
 
-    height = min (rest, 1);
+    height = MIN (rest, 1);
     view->status_area.height = height;
     rest -= height;
 
-    height = (ruler == RULER_NONE || view->hex_mode) ? 0 : 2;
-    height = min (rest, height);
+    height = (ruler == RULER_NONE || view->mode_flags.hex) ? 0 : 2;
+    height = MIN (rest, height);
     view->ruler_area.height = height;
     rest -= height;
 
@@ -317,7 +310,7 @@ mcview_compute_areas (mcview_t * view)
 /* --------------------------------------------------------------------------------------------- */
 
 void
-mcview_update_bytes_per_line (mcview_t * view)
+mcview_update_bytes_per_line (WView * view)
 {
     const screen_dimen cols = view->data_area.width;
     int bytes;
@@ -326,9 +319,8 @@ mcview_update_bytes_per_line (mcview_t * view)
         bytes = 4;
     else
         bytes = 4 * ((cols - 9) / ((cols <= 80) ? 17 : 18));
-#ifdef HAVE_ASSERT_H
-    assert (bytes != 0);
-#endif
+
+    g_assert (bytes != 0);
 
     view->bytes_per_line = bytes;
     view->dirty = mcview_max_dirt_limit + 1;    /* To force refresh */
@@ -337,7 +329,7 @@ mcview_update_bytes_per_line (mcview_t * view)
 /* --------------------------------------------------------------------------------------------- */
 
 void
-mcview_display_toggle_ruler (mcview_t * view)
+mcview_display_toggle_ruler (WView * view)
 {
     static const enum ruler_type next[3] =
     {
@@ -346,9 +338,8 @@ mcview_display_toggle_ruler (mcview_t * view)
         RULER_NONE
     };
 
-#ifdef HAVE_ASSERT_H
-    assert ((size_t) ruler < 3);
-#endif
+    g_assert ((size_t) ruler < 3);
+
     ruler = next[(size_t) ruler];
     mcview_compute_areas (view);
     view->dirty++;
@@ -357,7 +348,7 @@ mcview_display_toggle_ruler (mcview_t * view)
 /* --------------------------------------------------------------------------------------------- */
 
 void
-mcview_display_clean (mcview_t * view)
+mcview_display_clean (WView * view)
 {
     Widget *w = WIDGET (view);
 
@@ -370,7 +361,7 @@ mcview_display_clean (mcview_t * view)
 /* --------------------------------------------------------------------------------------------- */
 
 void
-mcview_display_ruler (mcview_t * view)
+mcview_display_ruler (WView * view)
 {
     static const char ruler_chars[] = "|----*----";
     const screen_dimen top = view->ruler_area.top;
