@@ -2,7 +2,7 @@
    Internal file viewer for the Midnight Commander
    Common finctions (used from some other mcviewer functions)
 
-   Copyright (C) 1994-2015
+   Copyright (C) 1994-2018
    Free Software Foundation, Inc.
 
    Written by:
@@ -35,7 +35,7 @@
 
 #include <config.h>
 
-#include <limits.h>
+#include <string.h>             /* memset() */
 #include <sys/types.h>
 
 #include "lib/global.h"
@@ -55,10 +55,6 @@
 
 /*** global variables ****************************************************************************/
 
-#define OFF_T_BITWIDTH (unsigned int) (sizeof (off_t) * CHAR_BIT - 1)
-const off_t INVALID_OFFSET = (off_t) (-1);
-const off_t OFFSETTYPE_MAX = ((off_t) 1 << (OFF_T_BITWIDTH - 1)) - 1;
-
 /*** file scope macro definitions ****************************************************************/
 
 /*** file scope type declarations ****************************************************************/
@@ -73,14 +69,14 @@ const off_t OFFSETTYPE_MAX = ((off_t) 1 << (OFF_T_BITWIDTH - 1)) - 1;
 /* --------------------------------------------------------------------------------------------- */
 
 void
-mcview_toggle_magic_mode (mcview_t * view)
+mcview_toggle_magic_mode (WView * view)
 {
     char *filename, *command;
     dir_list *dir;
     int *dir_idx;
 
-    mcview_altered_magic_flag = 1;
-    view->magic_mode = !view->magic_mode;
+    mcview_altered_flags.magic = TRUE;
+    view->mode_flags.magic = !view->mode_flags.magic;
 
     /* reinit view */
     filename = g_strdup (vfs_path_as_str (view->filename_vpath));
@@ -91,7 +87,7 @@ mcview_toggle_magic_mode (mcview_t * view)
     view->dir_idx = NULL;
     mcview_done (view);
     mcview_init (view);
-    mcview_load (view, command, filename, 0);
+    mcview_load (view, command, filename, 0, 0, 0);
     view->dir = dir;
     view->dir_idx = dir_idx;
     g_free (filename);
@@ -104,9 +100,9 @@ mcview_toggle_magic_mode (mcview_t * view)
 /* --------------------------------------------------------------------------------------------- */
 
 void
-mcview_toggle_wrap_mode (mcview_t * view)
+mcview_toggle_wrap_mode (WView * view)
 {
-    view->text_wrap_mode = !view->text_wrap_mode;
+    view->mode_flags.wrap = !view->mode_flags.wrap;
     view->dpy_wrap_dirty = TRUE;
     view->dpy_bbar_dirty = TRUE;
     view->dirty++;
@@ -115,10 +111,10 @@ mcview_toggle_wrap_mode (mcview_t * view)
 /* --------------------------------------------------------------------------------------------- */
 
 void
-mcview_toggle_nroff_mode (mcview_t * view)
+mcview_toggle_nroff_mode (WView * view)
 {
-    view->text_nroff_mode = !view->text_nroff_mode;
-    mcview_altered_nroff_flag = 1;
+    view->mode_flags.nroff = !view->mode_flags.nroff;
+    mcview_altered_flags.nroff = TRUE;
     view->dpy_wrap_dirty = TRUE;
     view->dpy_bbar_dirty = TRUE;
     view->dirty++;
@@ -127,11 +123,11 @@ mcview_toggle_nroff_mode (mcview_t * view)
 /* --------------------------------------------------------------------------------------------- */
 
 void
-mcview_toggle_hex_mode (mcview_t * view)
+mcview_toggle_hex_mode (WView * view)
 {
-    view->hex_mode = !view->hex_mode;
+    view->mode_flags.hex = !view->mode_flags.hex;
 
-    if (view->hex_mode)
+    if (view->mode_flags.hex)
     {
         view->hex_cursor = view->dpy_start;
         view->dpy_start = mcview_offset_rounddown (view->dpy_start, view->bytes_per_line);
@@ -143,7 +139,7 @@ mcview_toggle_hex_mode (mcview_t * view)
         view->hex_cursor = view->dpy_start;
         widget_want_cursor (WIDGET (view), FALSE);
     }
-    mcview_altered_hex_mode = 1;
+    mcview_altered_flags.hex = TRUE;
     view->dpy_paragraph_skip_lines = 0;
     view->dpy_wrap_dirty = TRUE;
     view->dpy_bbar_dirty = TRUE;
@@ -153,7 +149,7 @@ mcview_toggle_hex_mode (mcview_t * view)
 /* --------------------------------------------------------------------------------------------- */
 
 void
-mcview_init (mcview_t * view)
+mcview_init (WView * view)
 {
     size_t i;
 
@@ -205,27 +201,22 @@ mcview_init (mcview_t * view)
 /* --------------------------------------------------------------------------------------------- */
 
 void
-mcview_done (mcview_t * view)
+mcview_done (WView * view)
 {
     /* Save current file position */
     if (mcview_remember_file_position && view->filename_vpath != NULL)
     {
         save_file_position (view->filename_vpath, -1, 0,
-                            view->hex_mode ? view->hex_cursor : view->dpy_start,
+                            view->mode_flags.hex ? view->hex_cursor : view->dpy_start,
                             view->saved_bookmarks);
         view->saved_bookmarks = NULL;
     }
 
     /* Write back the global viewer mode */
-    mcview_default_hex_mode = view->hex_mode;
-    mcview_default_nroff_flag = view->text_nroff_mode;
-    mcview_default_magic_flag = view->magic_mode;
-    mcview_global_wrap_mode = view->text_wrap_mode;
+    mcview_global_flags = view->mode_flags;
 
     /* Free memory used by the viewer */
-
     /* view->widget needs no destructor */
-
     vfs_path_free (view->filename_vpath);
     view->filename_vpath = NULL;
     vfs_path_free (view->workdir_vpath);
@@ -235,7 +226,8 @@ mcview_done (mcview_t * view)
     mcview_close_datasource (view);
     /* the growing buffer is freed with the datasource */
 
-    coord_cache_free (view->coord_cache), view->coord_cache = NULL;
+    coord_cache_free (view->coord_cache);
+    view->coord_cache = NULL;
 
     if (view->converter == INVALID_CONV)
         view->converter = str_cnv_from_term;
@@ -267,7 +259,7 @@ mcview_done (mcview_t * view)
 /* --------------------------------------------------------------------------------------------- */
 
 void
-mcview_set_codeset (mcview_t * view)
+mcview_set_codeset (WView * view)
 {
 #ifdef HAVE_CHARSET
     const char *cp_id = NULL;
@@ -298,7 +290,7 @@ mcview_set_codeset (mcview_t * view)
 
 #ifdef HAVE_CHARSET
 void
-mcview_select_encoding (mcview_t * view)
+mcview_select_encoding (WView * view)
 {
     if (do_select_codepage ())
         mcview_set_codeset (view);
@@ -308,7 +300,7 @@ mcview_select_encoding (mcview_t * view)
 /* --------------------------------------------------------------------------------------------- */
 
 void
-mcview_show_error (mcview_t * view, const char *msg)
+mcview_show_error (WView * view, const char *msg)
 {
     if (mcview_is_in_panel (view))
         mcview_set_datasource_string (view, msg);
@@ -322,7 +314,7 @@ mcview_show_error (mcview_t * view, const char *msg)
  */
 
 off_t
-mcview_bol (mcview_t * view, off_t current, off_t limit)
+mcview_bol (WView * view, off_t current, off_t limit)
 {
     int c;
     off_t filesize;
@@ -357,16 +349,14 @@ mcview_bol (mcview_t * view, off_t current, off_t limit)
  */
 
 off_t
-mcview_eol (mcview_t * view, off_t current, off_t limit)
+mcview_eol (WView * view, off_t current)
 {
     int c, prev_ch = 0;
-    off_t filesize;
-    filesize = mcview_get_filesize (view);
+
     if (current < 0)
         return 0;
-    if (current >= filesize)
-        return filesize;
-    while (current < filesize && current < limit)
+
+    while (TRUE)
     {
         if (!mcview_get_byte (view, current, &c))
             break;
@@ -390,7 +380,7 @@ mcview_eol (mcview_t * view, off_t current, off_t limit)
 char *
 mcview_get_title (const WDialog * h, size_t len)
 {
-    const mcview_t *view = (const mcview_t *) find_widget_type (h, mcview_callback);
+    const WView *view = (const WView *) find_widget_type (h, mcview_callback);
     const char *modified = view->hexedit_mode && (view->change_list != NULL) ? "(*) " : "    ";
     const char *file_label;
     const char *view_filename;
@@ -410,7 +400,7 @@ mcview_get_title (const WDialog * h, size_t len)
 /* --------------------------------------------------------------------------------------------- */
 
 int
-mcview_calc_percent (mcview_t * view, off_t p)
+mcview_calc_percent (WView * view, off_t p)
 {
     const screen_dimen right = view->status_area.left + view->status_area.width;
     const screen_dimen height = view->status_area.height;
@@ -423,7 +413,7 @@ mcview_calc_percent (mcview_t * view, off_t p)
         return (-1);
 
     filesize = mcview_get_filesize (view);
-    if (view->hex_mode && filesize > 0)
+    if (view->mode_flags.hex && filesize > 0)
     {
         /* p can't be beyond the last char, only over that. Compensate for this. */
         filesize--;
@@ -440,3 +430,11 @@ mcview_calc_percent (mcview_t * view, off_t p)
 }
 
 /* --------------------------------------------------------------------------------------------- */
+
+void
+mcview_clear_mode_flags (mcview_mode_flags_t * flags)
+{
+    memset (flags, 0, sizeof (*flags));
+}
+
+/* --------------------------------------------------------------------------------------------- */
